@@ -29,7 +29,8 @@ def get_backend(tensor) -> "AbstractBackend":
     if _result is not None:
         return _result
 
-    for framework_name, backend in list(_loaded_backends.items()):
+    previously_loaded_backends = list(_loaded_backends.items())
+    for _framework_name, backend in previously_loaded_backends:
         if backend.is_appropriate_type(tensor):
             _type2backend[_type] = backend
             return backend
@@ -42,10 +43,12 @@ def get_backend(tensor) -> "AbstractBackend":
         backends += backend.__subclasses__()
         backend_subclasses.append(backend)
 
+    # handles modification of _loaded_backends from other thread, see #391
+    prev_backend_names = [x for x, _ in previously_loaded_backends]
     for BackendSubclass in backend_subclasses:
         if _debug_importing:
             print("Testing for subclass of ", BackendSubclass)
-        if BackendSubclass.framework_name not in _loaded_backends:
+        if BackendSubclass.framework_name not in prev_backend_names:
             # check that module was already imported. Otherwise it can't be imported
             if BackendSubclass.framework_name in sys.modules:
                 if _debug_importing:
@@ -56,7 +59,7 @@ def get_backend(tensor) -> "AbstractBackend":
                     _type2backend[_type] = backend
                     return backend
 
-    raise RuntimeError("Tensor type unknown to einops {}".format(type(tensor)))
+    raise RuntimeError(f"Tensor type unknown to einops {type(tensor)}")
 
 
 class AbstractBackend:
@@ -77,7 +80,8 @@ class AbstractBackend:
     def create_symbol(self, shape):
         raise NotImplementedError("framework doesn't support symbolic computations")
 
-    def eval_symbol(self, symbol, input_dict):
+    def eval_symbol(self, symbol, symbol_value_pairs):
+        # symbol-value pairs is list[tuple[symbol, value-tensor]]
         raise NotImplementedError("framework doesn't support symbolic computations")
 
     def arange(self, start, stop):
@@ -128,7 +132,7 @@ class AbstractBackend:
         raise NotImplementedError("backend does not provide layers")
 
     def __repr__(self):
-        return "<einops backend for {}>".format(self.framework_name)
+        return f"<einops backend for {self.framework_name}>"
 
     def einsum(self, pattern, *x):
         raise NotImplementedError("backend does not support einsum")
@@ -196,7 +200,7 @@ class JaxBackend(NumpyBackend):
     framework_name = "jax"
 
     def __init__(self):
-        super(JaxBackend, self).__init__()
+        super().__init__()
         self.onp = self.np
 
         import jax.numpy
@@ -247,7 +251,7 @@ class TorchBackend(AbstractBackend):
             return x.mean(dim=reduced_axes)
         elif operation in ("any", "all", "prod"):
             # pytorch supports reducing only one operation at a time
-            for i in list(sorted(reduced_axes))[::-1]:
+            for i in sorted(reduced_axes)[::-1]:
                 x = getattr(x, operation)(dim=i)
             return x
         else:
@@ -326,57 +330,6 @@ class CupyBackend(AbstractBackend):
         return self.cupy.einsum(pattern, *x)
 
 
-class ChainerBackend(AbstractBackend):
-    framework_name = "chainer"
-
-    def __init__(self):
-        import chainer
-        import numpy
-
-        self.numpy = numpy
-        self.chainer = chainer
-
-    def is_appropriate_type(self, tensor):
-        return isinstance(tensor, self.chainer.Variable)
-
-    def from_numpy(self, x):
-        return self.chainer.Variable(x.astype("float32"))
-
-    def to_numpy(self, x):
-        if isinstance(x, self.chainer.Variable):
-            x = x.data
-        return x
-
-    def arange(self, start, stop):
-        return self.numpy.arange(start, stop)
-
-    def reduce(self, x, operation, axes):
-        return getattr(self.chainer.functions, operation)(x, axis=axes)
-
-    def stack_on_zeroth_dimension(self, tensors: list):
-        return self.chainer.functions.stack(tensors)
-
-    def tile(self, x, repeats):
-        return self.chainer.functions.tile(x, repeats)
-
-    def concat(self, tensors, axis: int):
-        return self.chainer.functions.concat(tensors, axis=axis)
-
-    def add_axis(self, x, new_position):
-        return self.chainer.functions.expand_dims(x, new_position)
-
-    def is_float_type(self, x):
-        return x.dtype in ("float16", "float32", "float64", "float128", "bfloat16")
-
-    def layers(self):
-        from .layers import chainer
-
-        return chainer
-
-    def einsum(self, pattern, *x):
-        return self.chainer.functions.einsum(pattern, *x)
-
-
 class HashableTuple:
     """Overcomes non-hashability of symbolic elements"""
 
@@ -384,8 +337,7 @@ class HashableTuple:
         self.elements = elements
 
     def __iter__(self):
-        for x in self.elements:
-            yield x
+        yield from self.elements
 
     def __len__(self):
         return len(self.elements)
@@ -482,9 +434,9 @@ class TFKerasBackend(AbstractBackend):
     def create_symbol(self, shape):
         return self.keras.Input(batch_shape=shape)
 
-    def eval_symbol(self, symbol, input_dict):
-        model = self.keras.models.Model([var for (var, _) in input_dict], symbol)
-        return model.predict_on_batch([val for (_, val) in input_dict])
+    def eval_symbol(self, symbol, symbol_value_pairs):
+        model = self.keras.models.Model([var for (var, _) in symbol_value_pairs], symbol)
+        return model.predict_on_batch([val for (_, val) in symbol_value_pairs])
 
     def arange(self, start, stop):
         return self.K.arange(start, stop)
@@ -602,7 +554,7 @@ class PaddleBackend(AbstractBackend):
         self.paddle = paddle
 
     def is_appropriate_type(self, tensor):
-        return isinstance(tensor, (self.paddle.Tensor, self.paddle.static.Variable))
+        return self.paddle.is_tensor(tensor)
 
     def from_numpy(self, x):
         tensor = self.paddle.to_tensor(x)
@@ -706,10 +658,107 @@ class TinygradBackend(AbstractBackend):
         return x.repeat(repeats)
 
     def concat(self, tensors, axis: int):
-        return tensors[0].cat(tensors[1:], axis) if len(tensors) > 1 else tensors[0]
+        return tensors[0].cat(*tensors[1:], dim=axis) if len(tensors) > 1 else tensors[0]
 
     def is_float_type(self, x):
         return self.tinygrad.dtypes.is_float(x.dtype)
 
     def einsum(self, pattern, *x):
         return self.tinygrad.Tensor.einsum(pattern, *x)
+
+
+class PyTensorBackend(AbstractBackend):
+    framework_name = "pytensor"
+
+    def __init__(self):
+        from pytensor import tensor
+
+        self.pt = tensor
+
+    def is_appropriate_type(self, tensor):
+        return isinstance(tensor, self.pt.TensorVariable)
+
+    def is_float_type(self, x):
+        return x.dtype in self.pt.type.float_dtypes
+
+    def from_numpy(self, x):
+        return self.pt.as_tensor(x)
+
+    def to_numpy(self, x):
+        return x.eval()  # Will only work if there are no symbolic inputs
+
+    def create_symbol(self, shape):
+        if not isinstance(shape, tuple | list):
+            shape = (shape,)
+        return self.pt.tensor(shape=shape)
+
+    def eval_symbol(self, symbol, symbol_value_pairs):
+        return symbol.eval(dict(symbol_value_pairs))
+
+    def arange(self, start, stop):
+        return self.pt.arange(start, stop)
+
+    def shape(self, x):
+        # use the static shape dimensions where known
+        return tuple(
+            static_dim if static_dim is not None else symbolic_dim
+            for static_dim, symbolic_dim in zip(x.type.shape, x.shape)
+        )
+
+    def stack_on_zeroth_dimension(self, tensors: list):
+        return self.pt.stack(tensors)
+
+    def tile(self, x, repeats):
+        return self.pt.tile(x, repeats)
+
+    def concat(self, tensors, axis: int):
+        return self.pt.concatenate(tensors, axis=axis)
+
+    def add_axis(self, x, new_position):
+        return self.pt.expand_dims(x, new_position)
+
+    def einsum(self, pattern, *x):
+        return self.pt.einsum(pattern, *x)
+
+
+class MLXBackend(AbstractBackend):
+    framework_name = "mlx"
+
+    def __init__(self):
+        import mlx.core as mx
+        import numpy as np
+
+        self.mx = mx
+        self.np = np
+
+    def is_appropriate_type(self, tensor):
+        return isinstance(tensor, self.mx.array)
+
+    def from_numpy(self, x):
+        return self.mx.array(x)
+
+    def to_numpy(self, x):
+        if x.dtype == self.mx.bfloat16:
+            x = x.astype(self.mx.float32)
+        return self.np.array(x)
+
+    def arange(self, start, stop):
+        return self.mx.arange(start, stop)
+
+    def stack_on_zeroth_dimension(self, tensors: list):
+        return self.mx.stack(tensors)
+
+    def add_axes(self, x, new_position):
+        return self.mx.expand_dims(x, new_position)
+
+    def tile(self, x, repeats):
+        return self.mx.tile(x, repeats)
+
+    def concat(self, tensors, axis: int):
+        return self.mx.concatenate(tensors, axis=axis)
+
+    def is_float_type(self, x):
+        return self.mx.issubdtype(x.dtype, self.mx.floating)
+
+    def einsum(self, pattern, *x):
+        return self.mx.einsum(pattern, *x)
